# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class Missing(dict):

    def __init__(self, field, existence=None, null_value=None):
        super(Missing, self).__init__()
        self.field = field
        self.existence = existence
        self.null_value = null_value
        self["missing"] = self._build_dict()

    def _build_dict(self):
        missing = {"field": self.field}
        if self.existence is not None:
            missing["existence"] = self.existence
        if self.null_value is not None:
            missing["null_value"] = self.null_value
        return missing



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class Bool(dict):

    def __init__(self):
        super(Bool, self).__init__()
        self._must = []
        self._must_not = []
        self._should = []
        self["bool"] = {}

    def bool(self, must=None, must_not=None, should=None):
        if must:
            self._must.append(must)
            self["bool"]["must"] = self._must
        if must_not:
            self._must_not.append(must_not)
            self["bool"]["must_not"] = self._must_not
        if should:
            self._should.append(should)
            self["bool"]["should"] = self._should
        return self

    def must(self, block):
        return self.bool(must=block)

    def must_not(self, block):
        return self.bool(must_not=block)

    def should(self, block):
        return self.bool(should=block)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

class ScriptScore(dict):

    def __init__(self, script, params=None, lang=None):
        super(ScriptScore, self).__init__()
        self.script = script
        self.params = params
        self.lang = lang
        self._build_dict()

    def _build_dict(self):
        self["script"] = self.script
        if self.params:
            self["params"] = self.params
        if self.lang:
            self["lang"] = self.lang


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

class Type(dict):

    def __init__(self, type_name):
        super(Type, self).__init__()
        self.type_name = type_name
        self["type"] = self._build_dict()

    def _build_dict(self):
        return {
            "value": self.type_name
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import json
from copy import deepcopy

from elasticsearch import Elasticsearch
from . import Filter, Bool, QueryBuilder


class QuerySet(object):

    def __init__(self, host, query=None, index=''):
        self._host = host
        self._index = index
        if isinstance(query, QueryBuilder):
            self._q = query
        else:
            self._q = QueryBuilder(query_string=query)
        self._wrappers = []
        self._count = None
        self._conn = None
        self._finalized_query = None
        # Caching
        self._cache = None
        self._retrieved = 0
        self._per_request = 10

    def _clone(self):
        query = deepcopy(self._q)
        clone = QuerySet(self._host, query=query, index=self._index)
        clone._conn = self._conn
        return clone

    @property
    def objects(self):
        return self._clone()

    @property
    def _query(self):
        return self._q._finalize_query()

    def filter(self, f, operator="and"):
        self._q.filter(f, operator=operator)
        return self

    def score(self, script_score, boost_mode="replace"):
        self._q.score(script_score, boost_mode=boost_mode)
        return self

    def only(self, fields):
        self._q.fields(fields)
        return self

    def order_by(self, sorting):
        self._q.sort(sorting)
        return self

    def wrappers(self, wrapper):
        self._wrappers.append(wrapper)
        return self

    def count(self):
        return self._count

    def next(self):
        """
        Provide iteration capabilities

        Use a small object cache for performance
        """
        if not self._cache:
            self._cache = self._get_results()
            self._retrieved += len(self._cache)

        # If we don't have any other data to return, we just
        # stop the iteration.
        if not self._cache:
            raise StopIteration()

        # Consuming the cache and updating the "cursor"
        return self._cache.pop(0)

    def _get_results(self):
        start = self._retrieved
        if self._count is None:
            # Always perform the first query since we don't know the count
            upper_limit = start + 1
        else:
            upper_limit = self._count
        if start < upper_limit:
            end = self._retrieved + self._per_request
            results = self[start:end]
            return results
        return []

    def __len__(self):
        return self.count()

    def __iter__(self):
        return self

    def __getitem__(self, val):
        """
        Override __getitem__ so we can activate our ES call when we try to slice
        """
        start = val.start
        end = val.stop
        results = self._search(start, end)
        for wrapper in self._wrappers:
            results = wrapper(results)
        return results

    def _search(self, start, end):
        conn = self._get_connection()
        pagination_kwargs = self._get_pagination_kwargs(start, end)
        raw_results = conn.search(index=self._index, body=self._query, **pagination_kwargs)
        self._count = self._get_result_count(raw_results)
        return raw_results["hits"]["hits"]

    def _get_result_count(self, results):
        return int(results["hits"]["total"])

    def _get_pagination_kwargs(self, start, end):
        size = end - start
        kwargs = {
            'from_': start,  # from is a reserved word, so we use 'from_'
            'size': size
        }
        return kwargs

    def _get_connection(self):
        if not self._conn:
            host_connection_info = self._parse_host_connection_info(self._host)
            self._conn = Elasticsearch(host_connection_info)
        return self._conn

    def _parse_host_connection_info(self, host):
        if isinstance(host, (str, unicode)):
            return [{"host": host}]
        if isinstance(host, dict):
            return [host]
        # Default to just using what was given to us
        return host



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class GeoDistance(dict):

    def __init__(self, coordinates, distance, field_name="location"):
        """
        Geo Distance Filter

        Acceptable Input Styles:
        GeoDistance({'lat': 40, 'lon': -70}, '10km')
        GeoDistance({'lat': 40, 'lng': -70}, '10km')
        GeoDistance([40, -70], '10km')
        GeoDistance("-70,40", '10km')
        """
        super(GeoDistance, self).__init__()
        self.coordinates = self._parse_coordinates(coordinates)
        self.distance = distance
        self.field_name = field_name
        self["geo_distance"] = self._build_dict()

    def _build_dict(self):
        geo_distance = {
            "distance": self.distance,
            self.field_name: {
                'lat': self.coordinates[0],
                'lon': self.coordinates[1]
            }
        }
        return geo_distance

    def _parse_coordinates(self, coordinates):
        if isinstance(coordinates, list):
            lat = coordinates[1]
            lon = coordinates[0]
        if isinstance(coordinates, dict):
            lat = coordinates.pop('lat')
            lon = coordinates.values()[0]
        if isinstance(coordinates, (str, unicode)):
            lat, lon = coordinates.split(",")
        return map(float, [lat, lon])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class Range(dict):

    def __init__(self, field_name, gte=None, gt=None, lte=None, lt=None):
        super(Range, self).__init__()
        self.field_name = field_name
        self.gte = gte
        self.gt = gt
        self.lte = lte
        self.lt = lt
        self["range"] = self._build_dict()

    def _build_dict(self):
        _range = {
                self.field_name: {}
            }
        if self.gte:
            _range[self.field_name]["gte"] = self.gte
        if self.gt:
            _range[self.field_name]["gt"] = self.gt
        if self.lte:
            _range[self.field_name]["lte"] = self.lte
        if self.lt:
            _range[self.field_name]["lt"] = self.lt
        return _range



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class Filter(dict):

    def __init__(self, operator="and"):
        super(Filter, self).__init__()
        self.operator = operator
        self[self.operator] = []

    def filter(self, new_filter):
        self[self.operator].append(new_filter)
        return self



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

class Terms(dict):

    def __init__(self, field_name, terms):
        super(Terms, self).__init__()
        self.field_name = field_name
        self.terms = terms
        self["terms"] = self._build_dict()

    def _build_dict(self):
        return {
            self.field_name: self.terms
        }


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from . import Filter
from copy import deepcopy
from pyeqs.dsl import MatchAll


class QueryBuilder(object):

    def __init__(self, query_string=None):
        super(QueryBuilder, self).__init__()
        self._query = {}
        self._query_dsl = {}
        self._query_string = query_string
        self._filtered = False
        self._filter_dsl = None
        self._scored = False
        self._score_dsl = None
        self._sorted = False
        self._sorting = None
        self._fields = []
        self._build_query()

    def _build_query(self):
        """
        Build the base query dictionary
        """
        if self._query_string:
            self._query_dsl = {"query_string": {"query": self._query_string}}
        else:
            self._query_dsl = MatchAll()

    def _build_filtered_query(self, f, operator):
        """
        Create the root of the filter tree
        """
        self._filtered = True
        if isinstance(f, Filter):
            filter_object = f
        else:
            filter_object = Filter(operator).filter(f)
        self._filter_dsl = filter_object

    def filter(self, f, operator="and"):
        """
        Add a filter to the query

        Takes a Filter object, or a filterable DSL object.
        """
        if self._filtered:
            self._filter_dsl.filter(f)
        else:
            self._build_filtered_query(f, operator)
        return self

    def _build_sorted_query(self):
        self._sorted = True
        self._sorting = []

    def sort(self, sorting):
        if not self._sorted:
            self._build_sorted_query()
        self._sorting.append(sorting)
        return self

    def score(self, script_score, boost_mode="replace"):
        self._scored = True
        self._score_dsl = {
            "function_score": {
                "script_score": script_score,
                "boost_mode": boost_mode
            }
        }
        return self

    def fields(self, fields):
        self._fields.append(fields)

    def _finalize_query(self):
        query = {
            "query": self._query_dsl
        }
        if self._filtered:
            filtered_query = {
                "query": {
                    "filtered": {
                        "filter": self._filter_dsl
                    }
                }
            }
            filtered_query["query"]["filtered"]["query"] = deepcopy(query["query"])
            query = filtered_query

        if self._scored:
            scored_query = {
                "query": self._score_dsl
            }
            scored_query["query"]["function_score"]["query"] = deepcopy(query["query"])
            query = scored_query

        if self._sorted:
            query["sort"] = self._sorting

        if self._fields:
            query["fields"] = self._fields
        return query



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class Sort(dict):

    def __init__(self, field, order="asc", mode=None, missing=None):
        super(Sort, self).__init__()
        self.field = field
        self.order = order
        self.mode = mode
        self.missing = missing
        self._build_dict()

    def _build_dict(self):
        sorting = {}
        sorting["order"] = self.order
        if self.mode:
            sorting["mode"] = self.mode
        if self.missing:
            sorting["missing"] = self.missing
        self[self.field] = sorting



# #!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import re
import os
from setuptools import setup, find_packages


def parse_requirements():
    """
    Rudimentary parser for the `requirements.txt` file

    We just want to separate regular packages from links to pass them to the
    `install_requires` and `dependency_links` params of the `setup()`
    function properly.
    """
    try:
        requirements = \
            map(str.strip, local_file('requirements.txt'))
    except IOError:
        raise RuntimeError("Couldn't find the `requirements.txt' file :(")

    links = []
    pkgs = []
    for req in requirements:
        if not req:
            continue
        if 'http:' in req or 'https:' in req:
            links.append(req)
            name, version = re.findall("\#egg=([^\-]+)-(.+$)", req)[0]
            pkgs.append('{0}=={1}'.format(name, version))
        else:
            pkgs.append(req)

    return pkgs, links

local_file = lambda f: \
    open(os.path.join(os.path.dirname(__file__), f)).readlines()

#install_requires, dependency_links = parse_requirements()

if __name__ == '__main__':

    packages = find_packages(exclude=['*tests*'])

    setup(
        name="pyeqs",
        license="MIT",
        version='0.6.2',
        description=u'Django Querysets-esque implementation for Elasticsearch',
        author=u'Andrew Gross',
        author_email=u'andrew.w.gross@gmail.com',
        include_package_data=True,
        url='https://github.com/yipit/pyeqs',
        packages=packages,
        classifiers=(
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT',
            'Natural Language :: English',
            'Operating System :: Microsoft',
            'Operating System :: POSIX',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
        )
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import


class MatchAll(dict):

    def __init__(self):
        super(MatchAll, self).__init__()
        self._build_dict()

    def _build_dict(self):
        self["match_all"] = {}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .geo import GeoDistance  # noqa
from .range import Range  # noqa
from .script_score import ScriptScore  # noqa
from .sort import Sort  # noqa
from .term import Term  # noqa
from .terms import Terms  # noqa
from .type import Type  # noqa
from .match_all import MatchAll  # noqa
from .missing import Missing

__all__ = (
    'GeoDistance',
    'MatchAll',
    'Missing',
    'Range',
    'ScriptScore',
    'Sort',
    'Term',
    'Terms',
    'Type',
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

class Term(dict):

    def __init__(self, field_name, term):
        super(Term, self).__init__()
        self.field_name = field_name
        self.term = term
        self["term"] = self._build_dict()

    def _build_dict(self):
        return {
            self.field_name: self.term
        }


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .filter import Filter  # noqa
from .bool import Bool  # noqa
from .query_builder import QueryBuilder  # noqa
from .queryset import QuerySet  # noqa


__all__ = (
    'Filter',
    'Bool',
    'Queryset',
    'QueryBuilder',
)
